#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
import abc
import glob
import pathlib
import re
from typing import Dict, Tuple

from .deliverer import Destination


class SpooledProduct(abc.ABC):
    """
    A SpooledProduct is something interesting enough to deliver to an end-user.
    SpooledProducts might be science products or auxilliary products, but they
    might just be some temporary thing that came out of some processing, never
    to be seen again.

    SpooledProducts have a type and a path. More specific SpooledProducts may
    have other specific metadata unto themselves.
    """

    def __init__(self, path: pathlib.Path):
        self.path = path
        self.current_file = None

    @abc.abstractmethod
    def deliver_to(self, destination: Destination):
        """
        Deliver sets up a double dispatch so we can get the product type into the
        method itself, and behave differently depending on the type of product
        being delivered.

        :param destination: The destination to deliver to
        """
        pass

    def __iter__(self):
        self.current_file = next(self.path.iterdir())
        return self.current_file

    def __next__(self):
        return next(self.current_file)


class ExecutionBlock(SpooledProduct):
    """
    An execution block is a collection of SDM and BDF files. There is also metadata
    about the execution block that may be interesting for the purposes of delivery
    eventually.
    """

    @property
    def eb_name(self):
        return self.path.absolute().name

    def deliver_to(self, destination: Destination):
        # let's use our directory name as the relative path
        destination.add_directory(self.path, self.eb_name)

    def __str__(self):
        return f"execution block {self.eb_name}"


class CalibratedMS(SpooledProduct):
    """
    A Calibrated Measurement Set is the end result of a restore capability request and is used as an input to imaging.
    Deliver all files in the request's products directory.
    """

    @property
    def cms_name(self):
        return self.path.absolute().name

    def deliver_to(self, destination: Destination):
        destination.add_directory(self.path, self.cms_name)

    def __str__(self):
        return f"calibrated MS {self.cms_name}"


# Future types of product that might be needed:
#
class Calibration(SpooledProduct):
    def __init__(self, path: pathlib.Path):
        super().__init__(path)
        self.metadata = self.parse_metadata()

    def deliver_to(self, destination: Destination):
        # TODO: Deliver tarred calibration products and weblog to correctly named directory according to the structure:
        #   <project-code>
        #   └── calibration_pipeline.<pipeline-id-1>.<pipeline-id-2>
        #       |-- <project-code><timestamp>.tar
        #       └── weblog.tgz

        # TODO: Implement tarring of relevant files into calibration products archive
        """
        List of all files typically found in a calibration spool dir:
           X - not relevant file (exclude from tar

           products/
           |-- 21A-409.sb39530507.eb39564922.59311.953874212966.ms.calapply.txt
           |-- 21A-409.sb39530507.eb39564922.59311.953874212966.ms.flagversions.tgz
           |-- casa_commands.log
           |-- casa_piperestorescript.py
           |-- casa_pipescript.py
           |-- flux.csv
           |-- [X] oussid.J0319+4130_bp.A_band.cont.I.alpha.error.fits
           |-- [X] oussid.J0319+4130_bp.A_band.cont.I.alpha.fits
           |-- [X] oussid.J0319+4130_bp.A_band.cont.I.mask.fits
           |-- [X] oussid.J0319+4130_bp.A_band.cont.I.pb.tt0.fits
           |-- [X] oussid.J0319+4130_bp.A_band.cont.I.tt0.fits
           |-- [X] oussid.J0319+4130_bp.A_band.cont.I.tt1.fits
           |-- [X] oussid.J0552+0313_ph.A_band.cont.I.alpha.error.fits
           |-- [X] oussid.J0552+0313_ph.A_band.cont.I.alpha.fits
           |-- [X] oussid.J0552+0313_ph.A_band.cont.I.mask.fits
           |-- [X] oussid.J0552+0313_ph.A_band.cont.I.pb.tt0.fits
           |-- [X] oussid.J0552+0313_ph.A_band.cont.I.tt0.fits
           |-- [X] oussid.J0552+0313_ph.A_band.cont.I.tt1.fits
           |-- [X] pipeline-20210409T164741/
           |-- pipeline_aquareport.xml
           |-- PPR.xml
           |-- [X] reprocess_flag_file.fits
           |-- unknown.hifv_cal.auxproducts.tgz
           |-- unknown.hifv_cal.pipeline_manifest.xml
           |-- unknown.hifv_cal.pprequest.xml
           |-- unknown.session_1.caltables.tgz
           `-- [X] weblog.tgz
        """
        # let's parse the metadata and make some mistakes
        project_code, pipeline_id = self.parse_metadata()

        # change to the pipeline directory
        pipedir = destination / project_code / f"calibration_pipeline.{pipeline_id}"

        # let's omit some junk
        with pipedir.excluding(".fits", ".tgz") as nofits:
            # let's deliver everything that isn't a fits or tgz file
            for file in self.path.iterdir():
                nofits.add_path_entry(file, file.name)

        # add the weblog manually because it was masked above
        weblog = self.path / "weblog.tgz"
        if weblog.exists():
            pipedir.add_file(weblog, "weblog.tgz")

    def parse_metadata(self) -> Tuple[str, str]:
        """
        Scan through the calibration products and collect information about the calibration

        :return: Dict of calibration metadata

        TODO: Collect metadata from calapply file name: project code, CASA pipeline run identification numbers
        """
        # There should only be a single `.ms.calapply.txt` file
        calapply_file, *_ = list(self.path.glob("**/*.ms.calapply.txt"))

        # The calapply file has a name like 21A-409.sb39530507.eb39564922.59311.953874212966.ms.calapply.txt
        # For our purposes, we need the project code (21A-409) and the pipeline identifier (59311.953874212966).
        #
        # This regex obtains the project code (before the first dot), and then we look for two dot-separated groups
        # before the expected ending (ms.calapply.txt). Then we glue the second and third items back together to
        # recreate the pipeline identifier.
        parts = re.search(
            r"^(?P<project_code>[^.]+).*\.(?P<mjd>[^.]+)\.(?P<time>[^.]+)\.ms\.calapply\.txt$",
            calapply_file.name,
        )
        return parts["project_code"], f"{parts['mjd']}.{parts['time']}"

    def __str__(self):
        # TODO: Add a uniquely defining piece of metadata to the str
        return f"calibration"


class Image(SpooledProduct):
    @property
    def img_name(self):
        return self.path.absolute().name

    def deliver_to(self, destination: Destination):
        destination.add_file(self.path, self.img_name)

    def __str__(self):
        return f"Image {self.img_name}"


class Weblog(SpooledProduct):
    @property
    def file_name(self):
        return self.path.absolute().name

    def deliver_to(self, destination: Destination):
        destination.add_file(self.path, self.file_name)

    def __str__(self):
        return f"weblog"


#
#
# class ProcessingRequest(SpooledProduct):
#     def deliver_to(self, destination: Destination):
#         raise NotImplementedError
#
#
# class OUS(SpooledProduct):
#     """
#     OUSes have a great deal of sub-products within them.
#     """
#
#     def __init__(self, id):
#         self.id = id
#         self._products = []
#
#     def add_product(self, product):
#         """
#         Add a product to this OUS.
#
#         :param product:  the product to add
#         """
#         self._products.append(product)
#
#     def deliver_to(self, destination: Destination):
#         # FIXME: example implementation; refine
#         # The trick here is to basically "cd" into the OUS directory and then
#         # proceed as normal. The subdirectory decorator will ensure that the
#         # products have the same structure, but inside the OUS directory
#         # instead of the level above.
#         ous_destination = SubdirectoryDecorator(destination, self.id)
#         for product in self._products:
#             product.deliver_to(ous_destination)
#
